/****************************************************************************
 ****************************************************************************/

#include "jwm.h"

static void DispatchBorderButtonEvent(XButtonEvent *event, ClientNode *np);

/****************************************************************************
 ****************************************************************************/
void HandleButtonPress(XButtonEvent *event) {
	ClientNode *np;

	np = FindClientByParent(event->window);
	if(np) {
		XRaiseWindow(display, np->parent);
		switch(event->button) {
		case Button1:
			DispatchBorderButtonEvent(event, np);
			break;
		case Button3:
			MoveClient(np);
			break;
		default:
			break;
		}
	} else if(event->window == trayWindow) {
		HandleTrayClick(event);
	} else if(event->window == rootWindow) {
		ShowRootMenu(event->x, event->y);
	} else {
		np = FindClientByWindow(event->window);
		if(np) {
			XRaiseWindow(display, np->parent);
			XAllowEvents(display, ReplayPointer, CurrentTime);
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleConfigureRequest(XConfigureRequestEvent *event) {
	XWindowChanges wc;
	ClientNode *np;
	int north, south, east, west;

	np = FindClientByWindow(event->window);
	if(np) {

		if(event->value_mask & CWWidth) {
			np->width = event->width;
		}
		if(event->value_mask & CWHeight) {
			np->height = event->height;
		}
		if(event->value_mask & CWX) {
			np->x = event->x;
		}
		if(event->value_mask & CWY) {
			np->y = event->y;
		}

		if(np->borderFlags & BORDER_OUTLINE) {
			east = BORDER_EAST;
			west = BORDER_WEST;
			south = BORDER_SOUTH;
			if(np->borderFlags & BORDER_TITLE) {
				north = BORDER_NORTH;
			} else {
				north = BORDER_SOUTH;
			}
		} else {
			north = 0;
			south = 0;
			east = 0;
			west = 0;
		}

		wc.stack_mode = event->detail;
		wc.sibling = event->above;
		wc.border_width = 0;

		wc.x = west;
		wc.y = north;
		wc.width = np->width;
		wc.height = np->height;
		XConfigureWindow(display, np->window, event->value_mask, &wc);

		wc.x = np->x;
		wc.y = np->y;
		wc.width = np->width + east + west;
		wc.height = np->height + north + south;
		XConfigureWindow(display, np->parent, event->value_mask, &wc);

	} else {

		wc.x = event->x;
		wc.y = event->y;
		wc.width = event->width;
		wc.height = event->height;

		XConfigureWindow(display, event->window, event->value_mask, &wc);

	}

}

/****************************************************************************
 ****************************************************************************/
void HandleEnterNotify(XCrossingEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np && np->isMapped && !np->isActive) {
		FocusClient(np);
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleExpose(XExposeEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np) {
		DrawBorder(np);
	} else if(event->window == trayWindow) {
		DrawTray();
	}
}

/****************************************************************************
 ****************************************************************************/
void HandlePropertyNotify(XPropertyEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np) {

		switch(event->atom) {
		case XA_WM_NAME:
			if(np->name) {
				XFree(np->name);
			}
			XFetchName(display, np->window, &np->name);
			break;
		case XA_WM_NORMAL_HINTS:
			ReadClientHints(np);
			break;
		default:
			break;
		}

		DrawBorder(np);
		DrawTray();
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleClientMessage(XClientMessageEvent *event) {
Debug("ClientMessage");
}

/****************************************************************************
 ****************************************************************************/
void HandleMotionNotify(XMotionEvent *event) {
	ClientNode *np;
	Cursor cur;
	BorderActionType action;

	np = FindClientByParent(event->window);
	if(np && (np->borderFlags & BORDER_OUTLINE)) {
		action = GetBorderActionType(np, event->x, event->y);
		if(np->borderAction != action) {
			np->borderAction = action;
			cur = GetFrameCursor(action);
			XDefineCursor(display, np->parent, cur);
			DrawBorderButtons(np, np->parent, np->parentGC);
		}
	}

}
/****************************************************************************
 ****************************************************************************/
void HandleShapeEvent(XShapeEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np) {
		SetShape(np);
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleColormapChange(XColormapEvent *event) {
	ClientNode *np;

	if(event->new) {
		np = FindClientByWindow(event->window);
		if(np) {
			np->colorMap = event->colormap;
			XInstallColormap(display, np->colorMap);
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleDestroyNotify(XDestroyWindowEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np) {
		RemoveClient(np);
	}
}

/****************************************************************************
 ****************************************************************************/
void DispatchBorderButtonEvent(XButtonEvent *event, ClientNode *np) {
	BorderActionType action;

	if(!(np->borderFlags & BORDER_OUTLINE)) {
		return;
	}

	action = GetBorderActionType(np, event->x, event->y);

	switch(action & 0x0F) {
	case BA_RESIZE:
		ResizeClient(np, action);
		break;
	case BA_MOVE:
		MoveClient(np);
		break;
	case BA_CLOSE:
		DeleteClient(np);
		break;
	case BA_MAXIMIZE:
		MaximizeClient(np);
		break;
	case BA_MINIMIZE:
		MinimizeClient(np);
		break;
	default:
		break;
	}

}

