/***************************************************************************
 ***************************************************************************/

#include "jwm.h"

const int TRAY_HEIGHT = 28;
static const int TRAY_ITEM_HEIGHT = 18;
static const int TRAY_START = 64;
static const int TRAY_TEXT_START = 26;

typedef struct TrayNode {
	ClientNode *client;
	struct TrayNode *next;
} TrayNode;

Window trayWindow;
static GC trayGC;
static Pixmap buffer; 
static GC bufferGC;
static TrayNode *nodes = NULL;
static unsigned int itemCount = 0;
static unsigned int trayStop;

static unsigned int GetItemWidth();

/***************************************************************************
 ***************************************************************************/
void CreateTray() {
	XSetWindowAttributes attr;

	itemCount = 0;
	nodes = NULL;

	attr.override_redirect = True;
	attr.event_mask = ButtonPressMask | ExposureMask;
	attr.background_pixel = colors[COLOR_BORDER];

	trayWindow = XCreateWindow(display, rootWindow, 0,
		rootHeight - TRAY_HEIGHT - 1, rootWidth - 1, TRAY_HEIGHT, 1,
		DefaultDepth(display, 0), InputOutput,
		DefaultVisual(display, 0),
		CWOverrideRedirect | CWBackPixel | CWEventMask, &attr);
	XMapRaised(display, trayWindow);

	SetDefaultCursor(trayWindow);

	trayGC = XCreateGC(display, trayWindow, 0, NULL);
	XSetFont(display, trayGC, trayFont->fid);

	buffer = XCreatePixmap(display, rootWindow,
		rootWidth - trayStop - 1, TRAY_HEIGHT, DefaultDepth(display, 0));
	bufferGC = XCreateGC(display, buffer, 0, 0);
	XSetFont(display, bufferGC, trayFont->fid);

	trayStop = XTextWidth(trayFont, " 00:00 MM", 9) + LOAD_WIDTH + 5;

}

/***************************************************************************
 ***************************************************************************/
void AddClientToTray(ClientNode *np) {
	TrayNode *tp;

	tp = Allocate(sizeof(TrayNode));
	tp->client = np;
	tp->next = nodes;
	nodes = tp;

	++itemCount;

	DrawTray();
}

/***************************************************************************
 ***************************************************************************/
void RemoveClientFromTray(ClientNode *np) {
	TrayNode *tp, *lp;

	lp = NULL;
	for(tp = nodes; tp; tp = tp->next) {
		if(tp->client == np) {
			if(lp) {
				lp->next = tp->next;
			} else {
				nodes = tp->next;
			}
			--itemCount;
			Release(tp);
			break;
		}
		lp = tp;
	}

	DrawTray();
}

/***************************************************************************
 ***************************************************************************/
void DestroyTray() {
	TrayNode *tp;

	while(nodes) {
		tp = nodes->next;
		Release(nodes);
		nodes = tp;
	}

	XFreeGC(display, trayGC);
	XFreeGC(display, bufferGC);
	XFreePixmap(display, buffer);

	XDestroyWindow(display, trayWindow);
}

/***************************************************************************
 ***************************************************************************/
void HandleTrayClick(XButtonEvent *event) {
	TrayNode *tp;
	int x, y;
	int index;
	unsigned int width;

	XRaiseWindow(display, trayWindow);
	XAllowEvents(display, ReplayPointer, CurrentTime);

	GetMousePosition(&x, &y);

	y = y - rootHeight + TRAY_HEIGHT + 1;
	x -= TRAY_START;

	width = GetItemWidth();

	if(y > 1) {

		if(x < 0) {
			ShowRootMenu(0, rootHeight - TRAY_HEIGHT);
			return;
		}

		index = x / (int)width;
		for(tp = nodes; tp && index >= 0; tp = tp->next, index--) {
			if(index == 0) {
				switch(event->button) {
				case Button1:
					RestoreClient(tp->client);
					FocusClient(tp->client);
					break;
				case Button3:
					MinimizeClient(tp->client);
					break;
				default:
					break;
				}
				DrawTray();
				break;
			}
		}

	}

}

/***************************************************************************
 ***************************************************************************/
void DrawTray() {
	XImage *image;
	TrayNode *np;
	unsigned int width, x;
	unsigned int remainder;
	long borderPixel, borderTextPixel;

	if(shouldExit) {
		return;
	}

	XSetForeground(display, bufferGC, colors[COLOR_BORDER]);
	XFillRectangle(display, buffer, bufferGC, 0, 0,
		rootWidth - trayStop - 1, TRAY_HEIGHT);
	XPutImage(display, buffer, bufferGC, startImage, 0, 0,
		2, 2, startImage->width, startImage->height);

	if(!itemCount) {
		UpdateTime();
		XCopyArea(display, buffer, trayWindow, trayGC,
			0, 0, rootWidth - trayStop - 1, TRAY_HEIGHT, 0, 0);
		return;
	}

	width = GetItemWidth();

	x = TRAY_START;
	for(np = nodes; np; np = np->next) {

		if(np->client->isActive) {
			borderPixel = colors[COLOR_ACTIVE_BORDER];
			borderTextPixel = colors[COLOR_ACTIVE_BORDER_TEXT];
		} else {
			borderPixel = colors[COLOR_BORDER];
			borderTextPixel = colors[COLOR_BORDER_TEXT];
		}

		XSetForeground(display, bufferGC, borderPixel);
		XFillRectangle(display, buffer, bufferGC, x, 2,
			width - 2, TRAY_HEIGHT - 4);

		XSetForeground(display, bufferGC, borderTextPixel);
		XSetBackground(display, bufferGC, borderPixel);

		if(np->client->isMapped) {
			image = trayMaximizedImage;
		} else {
			image = trayMinimizedImage;
		}
		XPutImage(display, buffer, bufferGC, image, 0, 0,
			x + 4, 6, image->width, image->height);

		if(np->client->name && np->client->name[0]) {
			XDrawImageString(display, buffer, bufferGC,
				x + TRAY_TEXT_START, 20, np->client->name,
				strlen(np->client->name));
		}
		XSetForeground(display, bufferGC, colors[COLOR_BORDER]);
		XFillRectangle(display, buffer, bufferGC,
			x + width - 1, 3, 2, TRAY_HEIGHT - 5);

		XSetForeground(display, bufferGC, colors[COLOR_BORDER_TEXT]);
		XDrawRectangle(display, buffer, bufferGC, x, 2,
			width - 2, TRAY_HEIGHT - 4);

		x += width;
	}

	remainder = (width * itemCount) - rootWidth - TRAY_START - trayStop;
	if(remainder) {
		XSetForeground(display, bufferGC, colors[COLOR_BORDER]);
		XFillRectangle(display, buffer, bufferGC,
			x, 3, remainder, TRAY_HEIGHT - 5);
	}

	XCopyArea(display, buffer, trayWindow, trayGC,
		0, 0, rootWidth - trayStop - 1, TRAY_HEIGHT, 0, 0);

	UpdateTime();

}

/***************************************************************************
 ***************************************************************************/
void UpdateTime() {
	static char tstring[80];
#ifdef SHOW_LOAD
	Pixmap loadDisplay;
#endif
	unsigned int len;
	time_t clock;
	struct tm *timeInfo;
	unsigned int x;

	clock = time(NULL);
	timeInfo = localtime(&clock);
	len = strftime(tstring, sizeof(tstring), "%I:%M %p", timeInfo);
	if(tstring[0] == '0') {
		for(x = 0; x < len; x++) {
			tstring[x] = tstring[x + 1];
		}
		--len;
	}
	x = rootWidth - XTextWidth(trayFont, tstring, len) - 4;

	XSetForeground(display, trayGC, colors[COLOR_BORDER]);
	XFillRectangle(display, trayWindow, trayGC,
		rootWidth - trayStop, 0,
		trayStop, TRAY_HEIGHT);

	XSetBackground(display, trayGC, colors[COLOR_BORDER]);
	XSetForeground(display, trayGC, colors[COLOR_BORDER_TEXT]);
	XDrawImageString(display, trayWindow, trayGC, x, 2 + 18,
		tstring, len);

#ifdef SHOW_LOAD
	XSetForeground(display, trayGC, colors[COLOR_BORDER_TEXT]);
	XDrawRectangle(display, trayWindow, trayGC, rootWidth - trayStop,
		2, LOAD_WIDTH + 1, TRAY_HEIGHT - 4);

	loadDisplay = GetLoadDisplay();
	XCopyArea(display, loadDisplay, trayWindow, trayGC, 0, 0,
		LOAD_WIDTH, TRAY_HEIGHT - 5, rootWidth - trayStop + 1, 3);
#endif

}

/***************************************************************************
 ***************************************************************************/
unsigned int GetItemWidth() {
	unsigned int width;

	if(!itemCount) {
		return rootWidth - TRAY_START - trayStop;
	}

	width = rootWidth - TRAY_START - trayStop;
	width /= itemCount;
	if(!width) {
		width = 1;
	}

	return width;
}

